"""Constants for pyliebherrhomeapi."""

# API endpoints and configuration
API_BASE_URL = "https://home-api.smartdevice.liebherr.com"
API_VERSION = "v1"
DEFAULT_TIMEOUT = 10

# Control names
CONTROL_TEMPERATURE = "temperature"
CONTROL_SUPERFROST = "superfrost"
CONTROL_SUPERCOOL = "supercool"
CONTROL_PRESENTATION_LIGHT = "presentationlight"
CONTROL_PARTY_MODE = "partymode"
CONTROL_NIGHT_MODE = "nightmode"
CONTROL_ICE_MAKER = "icemaker"
CONTROL_HYDRO_BREEZE = "hydrobreeze"
CONTROL_BIO_FRESH_PLUS = "biofreshplus"
CONTROL_AUTO_DOOR = "autodoor"

# Temperature units
UNIT_CELSIUS = "°C"
UNIT_FAHRENHEIT = "°F"

# Ice maker modes
ICE_MAKER_OFF = "OFF"
ICE_MAKER_ON = "ON"
ICE_MAKER_MAX_ICE = "MAX_ICE"

# HydroBreeze modes
HYDRO_BREEZE_OFF = "OFF"
HYDRO_BREEZE_LOW = "LOW"
HYDRO_BREEZE_MEDIUM = "MEDIUM"
HYDRO_BREEZE_HIGH = "HIGH"

# BioFreshPlus modes
BIO_FRESH_PLUS_ZERO_ZERO = "ZERO_ZERO"
BIO_FRESH_PLUS_ZERO_MINUS_TWO = "ZERO_MINUS_TWO"
BIO_FRESH_PLUS_MINUS_TWO_MINUS_TWO = "MINUS_TWO_MINUS_TWO"
BIO_FRESH_PLUS_MINUS_TWO_ZERO = "MINUS_TWO_ZERO"

# Device types
DEVICE_TYPE_FRIDGE = "FRIDGE"
DEVICE_TYPE_FREEZER = "FREEZER"
DEVICE_TYPE_COMBI = "COMBI"
DEVICE_TYPE_WINE = "WINE"

# Zone positions
ZONE_TOP = "top"
ZONE_MIDDLE = "middle"
ZONE_BOTTOM = "bottom"

# Auto door states
DOOR_CLOSED = "CLOSED"
DOOR_OPEN = "OPEN"
DOOR_MOVING = "MOVING"



"""Tests for Liebherr exceptions."""

import pytest

from pyliebherrhomeapi import (
    LiebherrAuthenticationError,
    LiebherrBadRequestError,
    LiebherrConnectionError,
    LiebherrError,
    LiebherrNotFoundError,
    LiebherrPreconditionFailedError,
    LiebherrServerError,
    LiebherrTimeoutError,
    LiebherrUnsupportedError,
)


class TestExceptions:
    """Tests for exception classes."""

    def test_liebherr_error_base(self) -> None:
        """Test base LiebherrError exception."""
        error = LiebherrError("Test error")
        assert str(error) == "Test error"
        assert isinstance(error, Exception)

    def test_authentication_error(self) -> None:
        """Test LiebherrAuthenticationError."""
        error = LiebherrAuthenticationError("Authentication failed")
        assert str(error) == "Authentication failed"
        assert isinstance(error, LiebherrError)

    def test_bad_request_error(self) -> None:
        """Test LiebherrBadRequestError."""
        error = LiebherrBadRequestError("Invalid data")
        assert str(error) == "Invalid data"
        assert isinstance(error, LiebherrError)

    def test_connection_error(self) -> None:
        """Test LiebherrConnectionError."""
        error = LiebherrConnectionError("Connection failed")
        assert str(error) == "Connection failed"
        assert isinstance(error, LiebherrError)

    def test_not_found_error(self) -> None:
        """Test LiebherrNotFoundError."""
        error = LiebherrNotFoundError("Device not found")
        assert str(error) == "Device not found"
        assert isinstance(error, LiebherrError)

    def test_precondition_failed_error(self) -> None:
        """Test LiebherrPreconditionFailedError."""
        error = LiebherrPreconditionFailedError("Device not onboarded")
        assert str(error) == "Device not onboarded"
        assert isinstance(error, LiebherrError)

    def test_server_error(self) -> None:
        """Test LiebherrServerError."""
        error = LiebherrServerError("Internal server error")
        assert str(error) == "Internal server error"
        assert isinstance(error, LiebherrError)

    def test_timeout_error(self) -> None:
        """Test LiebherrTimeoutError."""
        error = LiebherrTimeoutError("Request timed out")
        assert str(error) == "Request timed out"
        assert isinstance(error, LiebherrError)

    def test_unsupported_error(self) -> None:
        """Test LiebherrUnsupportedError."""
        error = LiebherrUnsupportedError("Operation not supported")
        assert str(error) == "Operation not supported"
        assert isinstance(error, LiebherrError)

    def test_exception_inheritance(self) -> None:
        """Test that all exceptions inherit from LiebherrError."""
        exceptions = [
            LiebherrAuthenticationError,
            LiebherrBadRequestError,
            LiebherrConnectionError,
            LiebherrNotFoundError,
            LiebherrPreconditionFailedError,
            LiebherrServerError,
            LiebherrTimeoutError,
            LiebherrUnsupportedError,
        ]

        for exc_class in exceptions:
            error = exc_class("Test")
            assert isinstance(error, LiebherrError)
            assert isinstance(error, Exception)

    def test_exception_raising(self) -> None:
        """Test that exceptions can be raised and caught."""
        with pytest.raises(LiebherrError):
            raise LiebherrError("Test")

        with pytest.raises(LiebherrAuthenticationError):
            raise LiebherrAuthenticationError("Auth failed")

        with pytest.raises(LiebherrConnectionError):
            raise LiebherrConnectionError("Connection failed")



"""Data models for pyliebherrhomeapi."""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import Any, TypeVar

_EnumT = TypeVar("_EnumT", bound=Enum)


def _coerce_enum(enum_cls: type[_EnumT], value: str | None) -> _EnumT | str | None:
    """Return enum member when possible, else the raw value.

    This prevents hard failures when the upstream API introduces new values.
    """

    if value is None:
        return None
    try:
        return enum_cls(value)
    except ValueError:
        return value


class DeviceType(str, Enum):
    """Device type enumeration."""

    FRIDGE = "FRIDGE"
    FREEZER = "FREEZER"
    COMBI = "COMBI"
    WINE = "WINE"


class TemperatureUnit(str, Enum):
    """Temperature unit enumeration."""

    CELSIUS = "°C"
    FAHRENHEIT = "°F"


class ZonePosition(str, Enum):
    """Zone position enumeration."""

    TOP = "top"
    MIDDLE = "middle"
    BOTTOM = "bottom"


class IceMakerMode(str, Enum):
    """Ice maker mode enumeration."""

    OFF = "OFF"
    ON = "ON"
    MAX_ICE = "MAX_ICE"


class HydroBreezeMode(str, Enum):
    """HydroBreeze mode enumeration."""

    OFF = "OFF"
    LOW = "LOW"
    MEDIUM = "MEDIUM"
    HIGH = "HIGH"


class BioFreshPlusMode(str, Enum):
    """BioFreshPlus mode enumeration."""

    ZERO_ZERO = "ZERO_ZERO"
    ZERO_MINUS_TWO = "ZERO_MINUS_TWO"
    MINUS_TWO_MINUS_TWO = "MINUS_TWO_MINUS_TWO"
    MINUS_TWO_ZERO = "MINUS_TWO_ZERO"


class DoorState(str, Enum):
    """Door state enumeration."""

    CLOSED = "CLOSED"
    OPEN = "OPEN"
    MOVING = "MOVING"


class ControlType(str, Enum):
    """Control type enumeration."""

    TEMPERATURE = "TemperatureControl"
    TOGGLE = "ToggleControl"
    AUTO_DOOR = "AutoDoorControl"
    ICE_MAKER = "IceMakerControl"
    HYDRO_BREEZE = "HydroBreezeControl"
    BIO_FRESH_PLUS = "BioFreshPlusControl"


@dataclass
class Device:
    """Liebherr device information."""

    device_id: str
    nickname: str | None = None
    device_type: DeviceType | str | None = None
    image_url: str | None = None
    device_name: str | None = None

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Device:
        """Create Device from API response."""
        return cls(
            device_id=data["deviceId"],
            nickname=data.get("nickname"),
            device_type=_coerce_enum(DeviceType, data.get("deviceType")),
            image_url=data.get("imageUrl"),
            device_name=data.get("deviceName"),
        )

    def is_fridge(self) -> bool:
        """Check if device is a fridge."""
        return self.device_type == DeviceType.FRIDGE

    def is_freezer(self) -> bool:
        """Check if device is a freezer."""
        return self.device_type == DeviceType.FREEZER

    def is_combi(self) -> bool:
        """Check if device is a combination fridge/freezer."""
        return self.device_type == DeviceType.COMBI

    def is_wine(self) -> bool:
        """Check if device is a wine cooler."""
        return self.device_type == DeviceType.WINE


@dataclass
class TemperatureControl:
    """Temperature control information."""

    name: str
    type: str
    zone_id: int
    zone_position: ZonePosition | str | None = None
    value: int | None = None
    target: int | None = None
    min: int | None = None
    max: int | None = None
    unit: TemperatureUnit | str | None = None

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> TemperatureControl:
        """Create TemperatureControl from API response."""
        return cls(
            name=data["name"],
            type=data["type"],
            zone_id=data["zoneId"],
            zone_position=_coerce_enum(ZonePosition, data.get("zonePosition")),
            value=data.get("value"),
            target=data.get("target"),
            min=data.get("min"),
            max=data.get("max"),
            unit=_coerce_enum(TemperatureUnit, data.get("unit")),
        )

    def validate_temperature(self, temp: int) -> bool:
        """Validate if temperature is within allowed range.

        Args:
            temp: Temperature value to validate.

        Returns:
            True if temperature is within min/max range, False otherwise.

        """
        if self.min is not None and temp < self.min:
            return False
        if self.max is not None and temp > self.max:
            return False
        return True


@dataclass
class ToggleControl:
    """Toggle control (SuperCool, SuperFrost, etc.)."""

    name: str
    type: str
    zone_id: int | None = None
    zone_position: ZonePosition | str | None = None
    value: bool | None = None

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> ToggleControl:
        """Create ToggleControl from API response."""
        return cls(
            name=data["name"],
            type=data["type"],
            zone_id=data.get("zoneId"),
            zone_position=_coerce_enum(ZonePosition, data.get("zonePosition")),
            value=data.get("value"),
        )


@dataclass
class AutoDoorControl:
    """Auto door control information."""

    name: str
    type: str
    zone_id: int
    zone_position: ZonePosition | str | None = None
    value: DoorState | str | None = None

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> AutoDoorControl:
        """Create AutoDoorControl from API response."""
        return cls(
            name=data["name"],
            type=data["type"],
            zone_id=data["zoneId"],
            zone_position=_coerce_enum(ZonePosition, data.get("zonePosition")),
            value=_coerce_enum(DoorState, data.get("value")),
        )


@dataclass
class IceMakerControl:
    """Ice maker control information."""

    name: str
    type: str
    zone_id: int
    zone_position: ZonePosition | str | None = None
    ice_maker_mode: IceMakerMode | str | None = None
    has_max_ice: bool | None = None

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> IceMakerControl:
        """Create IceMakerControl from API response."""
        return cls(
            name=data["name"],
            type=data["type"],
            zone_id=data["zoneId"],
            zone_position=_coerce_enum(ZonePosition, data.get("zonePosition")),
            ice_maker_mode=_coerce_enum(IceMakerMode, data.get("iceMakerMode")),
            has_max_ice=data.get("hasMaxIce"),
        )


@dataclass
class HydroBreezeControl:
    """HydroBreeze control information."""

    name: str
    type: str
    zone_id: int
    current_mode: HydroBreezeMode | str | None = None

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> HydroBreezeControl:
        """Create HydroBreezeControl from API response."""
        return cls(
            name=data["name"],
            type=data["type"],
            zone_id=data["zoneId"],
            current_mode=_coerce_enum(HydroBreezeMode, data.get("currentMode")),
        )


@dataclass
class BioFreshPlusControl:
    """BioFreshPlus control information."""

    name: str
    type: str
    zone_id: int
    current_mode: BioFreshPlusMode | str | None = None
    supported_modes: list[BioFreshPlusMode | str] = field(default_factory=list)
    temperature_unit: TemperatureUnit | str | None = None

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> BioFreshPlusControl:
        """Create BioFreshPlusControl from API response."""
        supported: list[BioFreshPlusMode | str] = []
        for mode in data.get("supportedModes", []):
            coerced = _coerce_enum(BioFreshPlusMode, mode)
            if coerced is not None:
                supported.append(coerced)
        return cls(
            name=data["name"],
            type=data["type"],
            zone_id=data["zoneId"],
            current_mode=_coerce_enum(BioFreshPlusMode, data.get("currentMode")),
            supported_modes=supported,
            temperature_unit=_coerce_enum(TemperatureUnit, data.get("temperatureUnit")),
        )


DeviceControl = (
    TemperatureControl
    | ToggleControl
    | AutoDoorControl
    | IceMakerControl
    | HydroBreezeControl
    | BioFreshPlusControl
)


@dataclass
class DeviceState:
    """Complete device state including info and all controls."""

    device: Device
    controls: list[DeviceControl] = field(default_factory=list)

    def get_temperature_controls(self) -> list[TemperatureControl]:
        """Get all temperature controls.

        Returns:
            List of temperature controls.

        """
        return [
            control
            for control in self.controls
            if isinstance(control, TemperatureControl)
        ]

    def get_toggle_controls(self) -> list[ToggleControl]:
        """Get all toggle controls.

        Returns:
            List of toggle controls.

        """
        return [
            control for control in self.controls if isinstance(control, ToggleControl)
        ]

    def get_auto_door_controls(self) -> list[AutoDoorControl]:
        """Get all auto door controls.

        Returns:
            List of auto door controls.

        """
        return [
            control for control in self.controls if isinstance(control, AutoDoorControl)
        ]

    def get_ice_maker_controls(self) -> list[IceMakerControl]:
        """Get all ice maker controls.

        Returns:
            List of ice maker controls.

        """
        return [
            control for control in self.controls if isinstance(control, IceMakerControl)
        ]

    def get_hydro_breeze_controls(self) -> list[HydroBreezeControl]:
        """Get all HydroBreeze controls.

        Returns:
            List of HydroBreeze controls.

        """
        return [
            control
            for control in self.controls
            if isinstance(control, HydroBreezeControl)
        ]

    def get_biofresh_plus_controls(self) -> list[BioFreshPlusControl]:
        """Get all BioFreshPlus controls.

        Returns:
            List of BioFreshPlus controls.

        """
        return [
            control
            for control in self.controls
            if isinstance(control, BioFreshPlusControl)
        ]

    def get_control_by_name(self, name: str) -> DeviceControl | None:
        """Get control by name.

        Args:
            name: Control name to search for.

        Returns:
            Control with matching name, or None if not found.

        """
        for control in self.controls:
            if control.name == name:
                return control
        return None

    def get_controls_by_zone(self, zone_id: int) -> list[DeviceControl]:
        """Get all controls for a specific zone.

        Args:
            zone_id: Zone ID to filter by.

        Returns:
            List of controls for the specified zone.

        """
        zone_controls: list[DeviceControl] = []
        for control in self.controls:
            if isinstance(control, TemperatureControl) and control.zone_id == zone_id:
                zone_controls.append(control)
            elif isinstance(control, ToggleControl) and control.zone_id == zone_id:
                zone_controls.append(control)
            elif isinstance(control, AutoDoorControl) and control.zone_id == zone_id:
                zone_controls.append(control)
            elif isinstance(control, IceMakerControl) and control.zone_id == zone_id:
                zone_controls.append(control)
            elif isinstance(control, HydroBreezeControl) and control.zone_id == zone_id:
                zone_controls.append(control)
            elif (
                isinstance(control, BioFreshPlusControl) and control.zone_id == zone_id
            ):
                zone_controls.append(control)
        return zone_controls


def parse_control(data: dict[str, Any]) -> DeviceControl:
    """Parse device control from API response."""
    control_type = data.get("type")

    if control_type == ControlType.TEMPERATURE.value:
        return TemperatureControl.from_dict(data)
    if control_type == ControlType.TOGGLE.value:
        return ToggleControl.from_dict(data)
    if control_type == ControlType.AUTO_DOOR.value:
        return AutoDoorControl.from_dict(data)
    if control_type == ControlType.ICE_MAKER.value:
        return IceMakerControl.from_dict(data)
    if control_type == ControlType.HYDRO_BREEZE.value:
        return HydroBreezeControl.from_dict(data)
    if control_type == ControlType.BIO_FRESH_PLUS.value:
        return BioFreshPlusControl.from_dict(data)

    # Fallback to ToggleControl for unknown types
    return ToggleControl.from_dict(data)



"""Tests for Liebherr client."""
# pylint: disable=redefined-outer-name, protected-access

import importlib
from importlib.metadata import PackageNotFoundError
from typing import Any
from unittest.mock import AsyncMock, MagicMock, patch

import aiohttp
import pytest
from aiohttp.client_exceptions import ContentTypeError

from pyliebherrhomeapi import (
    BioFreshPlusMode,
    HydroBreezeMode,
    IceMakerMode,
    LiebherrAuthenticationError,
    LiebherrBadRequestError,
    LiebherrClient,
    LiebherrConnectionError,
    LiebherrNotFoundError,
    LiebherrPreconditionFailedError,
    LiebherrServerError,
    LiebherrTimeoutError,
    LiebherrUnsupportedError,
    TemperatureUnit,
)
from pyliebherrhomeapi.client import _get_version

API_KEY = "test-api-key"
DEVICE_ID = "12.345.678.9"


@pytest.fixture
def mock_response() -> MagicMock:
    """Create a mock response."""
    response = MagicMock()
    response.__aenter__ = AsyncMock(return_value=response)
    response.__aexit__ = AsyncMock(return_value=None)
    return response


@pytest.fixture
def mock_session(mock_response: MagicMock) -> MagicMock:
    """Create a mock aiohttp session."""
    session = MagicMock(spec=aiohttp.ClientSession)
    session.request = MagicMock(return_value=mock_response)
    return session


@pytest.fixture
def client(mock_session: MagicMock) -> LiebherrClient:
    """Create a test client."""
    return LiebherrClient(api_key=API_KEY, session=mock_session)


class TestClientLifecycle:
    """Tests for client lifecycle and configuration."""

    async def test_initialization(self) -> None:
        """Test client initialization."""
        client = LiebherrClient(api_key=API_KEY)
        assert client is not None
        await client.close()

    async def test_context_manager(self) -> None:
        """Test client as context manager."""
        async with LiebherrClient(api_key=API_KEY) as client:
            assert client is not None

    async def test_custom_base_url(self) -> None:
        """Test client with custom base URL."""
        custom_url = "https://custom.api.com/"
        client = LiebherrClient(api_key=API_KEY, base_url=custom_url)
        assert client._base_url == "https://custom.api.com"
        await client.close()

    async def test_custom_timeout(self) -> None:
        """Test client with custom timeout."""
        custom_timeout = 30
        client = LiebherrClient(api_key=API_KEY, timeout=custom_timeout)
        assert client._timeout == custom_timeout
        await client.close()

    async def test_creates_own_session(self) -> None:
        """Test client creates its own session when none provided."""
        client = LiebherrClient(api_key=API_KEY)
        assert client._session is None
        assert client._own_session is True

        mock_response = MagicMock()
        mock_response.__aenter__ = AsyncMock(return_value=mock_response)
        mock_response.__aexit__ = AsyncMock(return_value=None)
        mock_response.status = 200
        mock_response.json = AsyncMock(return_value=[])

        with patch.object(aiohttp.ClientSession, "request", return_value=mock_response):
            await client.get_devices()

        assert client._session is not None
        await client.close()
        assert client._session is None

    async def test_close_with_provided_session(self, mock_session: MagicMock) -> None:
        """Test that close doesn't close a session that was provided."""
        client = LiebherrClient(api_key=API_KEY, session=mock_session)
        assert client._own_session is False

        await client.close()
        mock_session.close.assert_not_called()

    async def test_close_when_no_session(self) -> None:
        """Test close when no session was created."""
        client = LiebherrClient(api_key=API_KEY)
        assert client._session is None
        await client.close()
        assert client._session is None


class TestDeviceOperations:
    """Tests for device-related operations."""

    async def test_get_devices(
        self, client: LiebherrClient, mock_response: MagicMock
    ) -> None:
        """Test getting all devices."""
        mock_response.status = 200
        mock_response.json = AsyncMock(
            return_value=[
                {
                    "deviceId": DEVICE_ID,
                    "nickname": "Kitchen Fridge",
                    "deviceType": "FRIDGE",
                    "deviceName": "Test Fridge",
                }
            ]
        )

        devices = await client.get_devices()
        assert len(devices) == 1
        assert devices[0].device_id == DEVICE_ID
        assert devices[0].nickname == "Kitchen Fridge"

    async def test_get_device(
        self, client: LiebherrClient, mock_response: MagicMock
    ) -> None:
        """Test getting a specific device."""
        mock_response.status = 200
        mock_response.json = AsyncMock(
            return_value={
                "deviceId": DEVICE_ID,
                "nickname": "Kitchen Fridge",
                "deviceType": "FRIDGE",
            }
        )

        device = await client.get_device(DEVICE_ID)
        assert device.device_id == DEVICE_ID

    @pytest.mark.parametrize(
        ("response_data",),
        [
            ({"error": "not a list"},),
        ],
    )
    async def test_get_devices_edge_cases(
        self,
        client: LiebherrClient,
        mock_response: MagicMock,
        response_data: list[Any] | dict[str, Any],
    ) -> None:
        """Test get_devices with edge cases."""
        mock_response.status = 200
        mock_response.json = AsyncMock(return_value=response_data)

        with pytest.raises(LiebherrServerError):
            await client.get_devices()

    async def test_get_device_not_dict(
        self, client: LiebherrClient, mock_response: MagicMock
    ) -> None:
        """Test getting device with non-dict response."""
        mock_response.status = 200
        mock_response.json = AsyncMock(return_value=["not", "a", "dict"])

        with pytest.raises(LiebherrServerError):
            await client.get_device(DEVICE_ID)


class TestControlOperations:
    """Tests for control-related operations."""

    async def test_get_controls(
        self, client: LiebherrClient, mock_response: MagicMock
    ) -> None:
        """Test getting device controls."""
        mock_response.status = 200
        mock_response.json = AsyncMock(
            return_value=[
                {
                    "name": "temperature",
                    "type": "TemperatureControl",
                    "zoneId": 0,
                    "value": 4,
                    "target": 4,
                    "min": 2,
                    "max": 8,
                    "unit": "°C",
                }
            ]
        )

        controls = await client.get_controls(DEVICE_ID)
        assert len(controls) == 1

    async def test_get_control(
        self, client: LiebherrClient, mock_response: MagicMock
    ) -> None:
        """Test getting specific control by name."""
        mock_response.status = 200
        mock_response.json = AsyncMock(
            return_value=[
                {
                    "name": "temperature",
                    "type": "TemperatureControl",
                    "zoneId": 0,
                    "value": 4,
                    "target": 4,
                }
            ]
        )

        controls = await client.get_control(DEVICE_ID, "temperature")
        assert len(controls) == 1
        assert controls[0].name == "temperature"

    async def test_get_control_with_zone(
        self, client: LiebherrClient, mock_response: MagicMock
    ) -> None:
        """Test getting specific control by name with zone filter."""
        mock_response.status = 200
        mock_response.json = AsyncMock(
            return_value=[
                {
                    "name": "temperature",
                    "type": "TemperatureControl",
                    "zoneId": 1,
                    "value": 4,
                }
            ]
        )

        controls = await client.get_control(DEVICE_ID, "temperature", zone_id=1)
        assert len(controls) == 1

    @pytest.mark.parametrize(
        ("response_data",),
        [
            ({"error": "not a list"},),
        ],
    )
    async def test_get_controls_edge_cases(
        self,
        client: LiebherrClient,
        mock_response: MagicMock,
        response_data: list[Any] | dict[str, Any],
    ) -> None:
        """Test get_controls with edge cases."""
        mock_response.status = 200
        mock_response.json = AsyncMock(return_value=response_data)

        with pytest.raises(LiebherrServerError):
            await client.get_controls(DEVICE_ID)

    @pytest.mark.parametrize(
        ("response_data",),
        [
            ({"error": "not a list"},),
        ],
    )
    async def test_get_control_edge_cases(
        self,
        client: LiebherrClient,
        mock_response: MagicMock,
        response_data: list[Any] | dict[str, Any],
    ) -> None:
        """Test get_control with edge cases."""
        mock_response.status = 200
        mock_response.json = AsyncMock(return_value=response_data)

        with pytest.raises(LiebherrServerError):
            await client.get_control(DEVICE_ID, "temperature")


class TestSetterMethods:
    """Tests for all setter methods using parametrization."""

    @pytest.mark.parametrize(
        ("method_name", "kwargs"),
        [
            (
                "set_temperature",
                {"zone_id": 0, "target": 4, "unit": TemperatureUnit.CELSIUS},
            ),
            ("set_superfrost", {"zone_id": 0, "value": True}),
            ("set_supercool", {"zone_id": 0, "value": True}),
            ("set_presentation_light", {"target": 50}),
            ("set_ice_maker", {"zone_id": 0, "mode": IceMakerMode.MAX_ICE}),
            ("set_hydro_breeze", {"zone_id": 0, "mode": HydroBreezeMode.HIGH}),
            ("set_bio_fresh_plus", {"zone_id": 0, "mode": BioFreshPlusMode.ZERO_ZERO}),
            ("trigger_auto_door", {"zone_id": 0, "value": True}),
            ("set_party_mode", {"value": True}),
            ("set_night_mode", {"value": True}),
        ],
    )
    async def test_setter_methods(
        self,
        client: LiebherrClient,
        mock_response: MagicMock,
        method_name: str,
        kwargs: dict[str, Any],
    ) -> None:
        """Test all setter methods."""
        mock_response.status = 204

        method = getattr(client, method_name)
        await method(device_id=DEVICE_ID, **kwargs)
        # Should not raise


class TestConvenienceMethods:
    """Tests for convenience methods."""

    async def test_get_device_state(
        self, client: LiebherrClient, mock_response: MagicMock
    ) -> None:
        """Test getting complete device state."""
        mock_response.status = 200
        device_response = {
            "deviceId": DEVICE_ID,
            "nickname": "Kitchen Fridge",
            "deviceType": "FRIDGE",
        }
        controls_response = [
            {
                "name": "temperature",
                "type": "TemperatureControl",
                "zoneId": 0,
                "value": 4,
            }
        ]
        mock_response.json = AsyncMock(side_effect=[device_response, controls_response])

        state = await client.get_device_state(DEVICE_ID)
        assert state.device.device_id == DEVICE_ID
        assert len(state.controls) == 1

    async def test_refresh_device(
        self, client: LiebherrClient, mock_response: MagicMock
    ) -> None:
        """Test refresh_device (alias for get_device_state)."""
        mock_response.status = 200
        device_response = {"deviceId": DEVICE_ID, "deviceType": "FRIDGE"}
        controls_response: list[dict[str, Any]] = []
        mock_response.json = AsyncMock(side_effect=[device_response, controls_response])

        state = await client.refresh_device(DEVICE_ID)
        assert state.device.device_id == DEVICE_ID
        assert len(state.controls) == 0

    async def test_get_temperature_controls(
        self, client: LiebherrClient, mock_response: MagicMock
    ) -> None:
        """Test getting only temperature controls."""
        mock_response.status = 200
        mock_response.json = AsyncMock(
            return_value=[
                {
                    "name": "temperature",
                    "type": "TemperatureControl",
                    "zoneId": 0,
                    "value": 4,
                },
                {
                    "name": "superfrost",
                    "type": "ToggleControl",
                    "zoneId": 0,
                    "value": True,
                },
            ]
        )

        temp_controls = await client.get_temperature_controls(DEVICE_ID)
        assert len(temp_controls) == 1
        assert temp_controls[0].name == "temperature"


class TestErrorHandling:
    """Tests for error handling using parametrization."""

    @pytest.mark.parametrize(
        ("status", "response_data", "exception_class"),
        [
            (401, {"message": "Unauthorized"}, LiebherrAuthenticationError),
            (400, {"message": "Invalid data"}, LiebherrBadRequestError),
            (404, {"message": "Device not found"}, LiebherrNotFoundError),
            (412, {"message": "Device not onboarded"}, LiebherrPreconditionFailedError),
            (422, {"message": "Not supported"}, LiebherrUnsupportedError),
            (500, {"message": "Internal error"}, LiebherrServerError),
            (503, {"message": "Service unavailable"}, LiebherrConnectionError),
        ],
    )
    async def test_http_errors(
        self,
        client: LiebherrClient,
        mock_response: MagicMock,
        status: int,
        response_data: dict[str, str],
        exception_class: type[Exception],
    ) -> None:
        """Test HTTP error handling."""
        mock_response.status = status
        mock_response.json = AsyncMock(return_value=response_data)

        with pytest.raises(exception_class):
            await client.get_devices()

    async def test_timeout_error(self, mock_session: MagicMock) -> None:
        """Test timeout error."""
        mock_session.request.side_effect = TimeoutError("Request timed out")
        client = LiebherrClient(api_key=API_KEY, session=mock_session)

        with pytest.raises(LiebherrTimeoutError):
            await client.get_devices()

    async def test_connection_error(self, mock_session: MagicMock) -> None:
        """Test connection error."""
        mock_session.request.side_effect = aiohttp.ClientError("Connection failed")
        client = LiebherrClient(api_key=API_KEY, session=mock_session)

        with pytest.raises(LiebherrConnectionError):
            await client.get_devices()

    async def test_http_status_204(
        self, client: LiebherrClient, mock_response: MagicMock
    ) -> None:
        """Test handling of 204 No Content response."""
        mock_response.status = 204

        result = await client._request("POST", "test")
        assert result is None

    async def test_extract_message_content_type_error(
        self, client: LiebherrClient, mock_response: MagicMock
    ) -> None:
        """Fallback to response text when JSON parsing fails due to content type."""

        mock_response.status = 400
        mock_response.reason = "Bad Request"
        mock_response.text = AsyncMock(return_value="plain error")
        mock_response.json = AsyncMock(
            side_effect=ContentTypeError(MagicMock(), (), message="bad content")
        )

        with pytest.raises(LiebherrBadRequestError) as err:
            await client.get_devices()

        assert "plain error" in str(err.value)

    async def test_extract_message_returns_reason_when_not_dict(
        self, client: LiebherrClient, mock_response: MagicMock
    ) -> None:
        """Fallback to response reason when JSON is not a dict."""

        mock_response.status = 400
        mock_response.reason = "Bad Request"
        mock_response.json = AsyncMock(return_value=["oops"])

        with pytest.raises(LiebherrBadRequestError) as err:
            await client.get_devices()

        assert "Bad Request" in str(err.value)

    async def test_unexpected_json_format_raises_server_error(
        self, client: LiebherrClient, mock_response: MagicMock
    ) -> None:
        """Non-JSON 200 responses raise server error with context."""

        mock_response.status = 200
        mock_response.reason = "OK"
        mock_response.raise_for_status = MagicMock(return_value=None)
        mock_response.text = AsyncMock(return_value="not json")
        mock_response.json = AsyncMock(side_effect=ValueError("boom"))

        with pytest.raises(LiebherrServerError) as err:
            await client.get_devices()

        assert "Unexpected response format" in str(err.value)

    async def test_raise_for_status(
        self, client: LiebherrClient, mock_response: MagicMock
    ) -> None:
        """Test that unknown status codes call raise_for_status."""
        mock_response.status = 418
        mock_response.raise_for_status.side_effect = aiohttp.ClientResponseError(
            request_info=MagicMock(),
            history=(),
            status=418,
            message="I'm a teapot",
        )

        with pytest.raises(LiebherrConnectionError):
            await client.get_devices()


class TestVersionFallback:
    """Tests for version fallback handling."""

    def test_get_version_fallback(self, monkeypatch: pytest.MonkeyPatch) -> None:
        """Return fallback version when package metadata is missing."""

        monkeypatch.setattr(
            "pyliebherrhomeapi.client.version",
            MagicMock(side_effect=PackageNotFoundError()),
        )

        assert _get_version() == "0.0.0"

    def test_module_version_fallback(self, monkeypatch: pytest.MonkeyPatch) -> None:
        """Reload module to apply version fallback when metadata missing."""

        import pyliebherrhomeapi as module

        monkeypatch.setattr(
            "importlib.metadata.version",
            MagicMock(side_effect=PackageNotFoundError()),
        )

        reloaded = importlib.reload(module)

        assert reloaded.__version__ == "0.0.0"



"""Python library for Liebherr Home API."""

from importlib.metadata import PackageNotFoundError, version

from .client import LiebherrClient
from .exceptions import (
    LiebherrAuthenticationError,
    LiebherrBadRequestError,
    LiebherrConnectionError,
    LiebherrError,
    LiebherrNotFoundError,
    LiebherrPreconditionFailedError,
    LiebherrServerError,
    LiebherrTimeoutError,
    LiebherrUnsupportedError,
)
from .models import (
    AutoDoorControl,
    BioFreshPlusControl,
    BioFreshPlusMode,
    Device,
    DeviceControl,
    DeviceState,
    DeviceType,
    DoorState,
    HydroBreezeControl,
    HydroBreezeMode,
    IceMakerControl,
    IceMakerMode,
    TemperatureControl,
    TemperatureUnit,
    ToggleControl,
    ZonePosition,
)

try:
    __version__ = version("pyliebherrhomeapi")
except PackageNotFoundError:
    __version__ = "0.0.0"

__all__ = [
    # Client
    "LiebherrClient",
    # Exceptions
    "LiebherrAuthenticationError",
    "LiebherrBadRequestError",
    "LiebherrConnectionError",
    "LiebherrError",
    "LiebherrNotFoundError",
    "LiebherrPreconditionFailedError",
    "LiebherrServerError",
    "LiebherrTimeoutError",
    "LiebherrUnsupportedError",
    # Models
    "AutoDoorControl",
    "BioFreshPlusControl",
    "BioFreshPlusMode",
    "Device",
    "DeviceControl",
    "DeviceState",
    "DeviceType",
    "DoorState",
    "HydroBreezeControl",
    "HydroBreezeMode",
    "IceMakerControl",
    "IceMakerMode",
    "TemperatureControl",
    "TemperatureUnit",
    "ToggleControl",
    "ZonePosition",
]



"""Client for Liebherr Home API.

Terminology (from Liebherr SmartDevice HomeAPI documentation):
- device: The Liebherr appliance
- deviceId: The serial number of the appliance
- zone: A cooling/freezing zone in the device (min. 1 zone per device)
    - Zone 0 is the top zone
    - Zone numbers ascend from top to bottom
- base controls: Controls that apply to the whole device (e.g., Party Mode)
- zone controls: Controls that apply to a specific zone (e.g., Temperature)

Important Notes:
- Only appliances connected via SmartDevice app are accessible
- Zone controls always require a zone_id, even if device has only one zone
- Recommended polling interval: 30 seconds for controls
- API key is obtained from SmartDevice app (Settings -> Beta features -> HomeAPI)
- API key can only be copied once from the app!
"""

from __future__ import annotations

from importlib.metadata import PackageNotFoundError, version
from typing import Any

import aiohttp
from aiohttp import ContentTypeError

from .const import (
    API_BASE_URL,
    API_VERSION,
    CONTROL_AUTO_DOOR,
    CONTROL_BIO_FRESH_PLUS,
    CONTROL_HYDRO_BREEZE,
    CONTROL_ICE_MAKER,
    CONTROL_NIGHT_MODE,
    CONTROL_PARTY_MODE,
    CONTROL_PRESENTATION_LIGHT,
    CONTROL_SUPERCOOL,
    CONTROL_SUPERFROST,
    CONTROL_TEMPERATURE,
    DEFAULT_TIMEOUT,
)
from .exceptions import (
    LiebherrAuthenticationError,
    LiebherrBadRequestError,
    LiebherrConnectionError,
    LiebherrNotFoundError,
    LiebherrPreconditionFailedError,
    LiebherrServerError,
    LiebherrTimeoutError,
    LiebherrUnsupportedError,
)
from .models import (
    BioFreshPlusMode,
    Device,
    DeviceControl,
    DeviceState,
    HydroBreezeMode,
    IceMakerMode,
    TemperatureControl,
    TemperatureUnit,
    parse_control,
)


def _get_version() -> str:
    """Return installed package version with a safe fallback."""

    try:
        return version("pyliebherrhomeapi")
    except PackageNotFoundError:
        return "0.0.0"


class LiebherrClient:
    """Client for interacting with Liebherr Home API."""

    def __init__(
        self,
        api_key: str,
        session: aiohttp.ClientSession | None = None,
        timeout: int = DEFAULT_TIMEOUT,
        base_url: str = API_BASE_URL,
    ) -> None:
        """Initialize the Liebherr client.

        Args:
            api_key: API key for authentication.
            session: Optional aiohttp session. If not provided, new one created.
            timeout: Request timeout in seconds.
            base_url: Base URL for the API (default: production URL).

        """
        self._api_key = api_key
        self._session = session
        self._timeout = timeout
        self._base_url = base_url.rstrip("/")
        self._own_session = session is None
        self._user_agent = f"pyliebherrhomeapi/{_get_version()}"

    async def _get_session(self) -> aiohttp.ClientSession:
        """Get or create aiohttp session."""
        if self._session is None:
            self._session = aiohttp.ClientSession()
        return self._session

    async def _request(
        self,
        method: str,
        endpoint: str,
        json_data: dict[str, Any] | None = None,
        params: dict[str, Any] | None = None,
    ) -> dict[str, Any] | list[Any] | None:
        """Make an API request.

        Args:
            method: HTTP method (GET, POST, etc.).
            endpoint: API endpoint path.
            json_data: JSON data for POST requests.
            params: Query parameters.

        Returns:
            Response data as dict, list, or None for 204 responses.

        Raises:
            LiebherrAuthenticationError: If authentication fails.
            LiebherrBadRequestError: If invalid data is provided.
            LiebherrNotFoundError: If device is not reachable.
            LiebherrPreconditionFailedError: If precondition fails.
            LiebherrUnsupportedError: If operation is not supported.
            LiebherrServerError: If server returns 500 error.
            LiebherrConnectionError: If connection fails.
            LiebherrTimeoutError: If request times out.

        """
        url = f"{self._base_url}/{API_VERSION}/{endpoint}"
        headers = {
            "api-key": self._api_key,
            "User-Agent": self._user_agent,
        }
        session = await self._get_session()

        try:
            async with session.request(
                method,
                url,
                json=json_data,
                params=params,
                headers=headers,
                timeout=aiohttp.ClientTimeout(total=self._timeout),
            ) as response:
                if response.status == 204:
                    return None

                async def _safe_json() -> Any:
                    return await response.json()

                async def _extract_message() -> str:
                    try:
                        error_data = await _safe_json()
                        if isinstance(error_data, dict):
                            return str(error_data.get("message", "Unknown error"))
                    except (ContentTypeError, ValueError):
                        text = await response.text()
                        return text.strip() or response.reason or ""
                    return response.reason or ""

                if response.status == 401:
                    raise LiebherrAuthenticationError("Authentication failed")
                if response.status == 400:
                    msg = await _extract_message()
                    raise LiebherrBadRequestError(f"Invalid data provided: {msg}")
                if response.status == 404:
                    msg = await _extract_message()
                    raise LiebherrNotFoundError(f"Device is not reachable: {msg}")
                if response.status == 412:
                    msg = await _extract_message()
                    raise LiebherrPreconditionFailedError(f"Precondition failed: {msg}")
                if response.status == 422:
                    msg = await _extract_message()
                    raise LiebherrUnsupportedError(f"Operation not supported: {msg}")
                if response.status == 500:
                    msg = await _extract_message()
                    raise LiebherrServerError(f"Internal server error: {msg}")
                if response.status == 503:
                    msg = await _extract_message()
                    raise LiebherrConnectionError(
                        f"Internal service not reachable: {msg}"
                    )

                try:
                    response.raise_for_status()
                except aiohttp.ClientResponseError as err:
                    msg = err.message or response.reason or ""
                    raise LiebherrConnectionError(
                        f"HTTP {response.status}: {msg}"
                    ) from err
                try:
                    data: dict[str, Any] | list[Any] = await _safe_json()
                except (ContentTypeError, ValueError) as err:
                    msg = await _extract_message()
                    raise LiebherrServerError(
                        f"Unexpected response format ({response.status}): {msg}"
                    ) from err
                return data

        except (TimeoutError, aiohttp.ServerTimeoutError) as ex:
            raise LiebherrTimeoutError("Request timed out") from ex
        except aiohttp.ClientError as ex:
            raise LiebherrConnectionError(f"Connection error: {ex}") from ex

    async def close(self) -> None:
        """Close the client session."""
        if self._own_session and self._session:
            await self._session.close()
            self._session = None

    async def __aenter__(self) -> LiebherrClient:
        """Async context manager entry."""
        return self

    async def __aexit__(self, *args: Any) -> None:
        """Async context manager exit."""
        await self.close()

    # Device endpoints

    async def get_devices(self) -> list[Device]:
        """Get all connected devices.

        Returns:
            List of Device objects.

        Raises:
            LiebherrConnectionError: If connection fails.
            LiebherrTimeoutError: If request times out.

        """
        response = await self._request("GET", "devices")
        if not isinstance(response, list):
            raise LiebherrServerError("Unexpected response format for devices")
        return [Device.from_dict(device) for device in response]

    async def get_device(self, device_id: str) -> Device:
        """Get a specific device by ID.

        Args:
            device_id: The device ID (serial number).

        Returns:
            Device object.

        Raises:
            LiebherrNotFoundError: If device is not found.
            LiebherrConnectionError: If connection fails.
            LiebherrTimeoutError: If request times out.

        """
        response = await self._request("GET", f"devices/{device_id}")
        if not isinstance(response, dict):
            raise LiebherrServerError("Unexpected response format for device")
        return Device.from_dict(response)

    # Control endpoints

    async def get_controls(self, device_id: str) -> list[DeviceControl]:
        """Get all controls for a device.

        Args:
            device_id: The device ID (serial number).

        Returns:
            List of control objects.

        Raises:
            LiebherrNotFoundError: If device is not found.
            LiebherrConnectionError: If connection fails.
            LiebherrTimeoutError: If request times out.

        """
        response = await self._request("GET", f"devices/{device_id}/controls")
        if not isinstance(response, list):
            raise LiebherrServerError("Unexpected response format for controls")
        return [parse_control(control) for control in response]

    async def get_control(
        self,
        device_id: str,
        control_name: str,
        zone_id: int | None = None,
    ) -> list[DeviceControl]:
        """Get specific control by name.

        Args:
            device_id: The device ID (serial number).
            control_name: Name of the control.
            zone_id: Optional zone ID for filtering.

        Returns:
            List of control objects matching the criteria.

        Raises:
            LiebherrNotFoundError: If device or control is not found.
            LiebherrConnectionError: If connection fails.
            LiebherrTimeoutError: If request times out.

        """
        params = {"zoneId": zone_id} if zone_id is not None else None
        response = await self._request(
            "GET",
            f"devices/{device_id}/controls/{control_name}",
            params=params,
        )
        if not isinstance(response, list):
            raise LiebherrServerError("Unexpected response format for control")
        return [parse_control(control) for control in response]

    # Temperature control

    async def set_temperature(
        self,
        device_id: str,
        zone_id: int,
        target: int,
        unit: TemperatureUnit = TemperatureUnit.CELSIUS,
    ) -> None:
        """Set temperature for a zone.

        Args:
            device_id: The device ID (serial number).
            zone_id: The zone ID.
            target: Target temperature.
            unit: Temperature unit (default: Celsius).

        Raises:
            LiebherrBadRequestError: If invalid data is provided.
            LiebherrNotFoundError: If device is not reachable.
            LiebherrPreconditionFailedError: If device not onboarded.
            LiebherrConnectionError: If connection fails.
            LiebherrTimeoutError: If request times out.

        """
        await self._request(
            "POST",
            f"devices/{device_id}/controls/{CONTROL_TEMPERATURE}",
            json_data={
                "zoneId": zone_id,
                "target": target,
                "unit": unit.value,
            },
        )

    # Toggle controls (SuperFrost, SuperCool, etc.)

    async def set_superfrost(self, device_id: str, zone_id: int, value: bool) -> None:
        """Set SuperFrost mode.

        Args:
            device_id: The device ID (serial number).
            zone_id: The zone ID.
            value: True to enable, False to disable.

        Raises:
            LiebherrBadRequestError: If invalid data is provided.
            LiebherrNotFoundError: If device is not reachable.
            LiebherrPreconditionFailedError: If device not onboarded.
            LiebherrConnectionError: If connection fails.
            LiebherrTimeoutError: If request times out.

        """
        await self._request(
            "POST",
            f"devices/{device_id}/controls/{CONTROL_SUPERFROST}",
            json_data={"zoneId": zone_id, "value": value},
        )

    async def set_supercool(self, device_id: str, zone_id: int, value: bool) -> None:
        """Set SuperCool mode.

        Args:
            device_id: The device ID (serial number).
            zone_id: The zone ID.
            value: True to enable, False to disable.

        Raises:
            LiebherrBadRequestError: If invalid data is provided.
            LiebherrNotFoundError: If device is not reachable.
            LiebherrPreconditionFailedError: If device not onboarded.
            LiebherrConnectionError: If connection fails.
            LiebherrTimeoutError: If request times out.

        """
        await self._request(
            "POST",
            f"devices/{device_id}/controls/{CONTROL_SUPERCOOL}",
            json_data={"zoneId": zone_id, "value": value},
        )

    async def set_party_mode(self, device_id: str, value: bool) -> None:
        """Set PartyMode.

        Args:
            device_id: The device ID (serial number).
            value: True to enable, False to disable.

        Raises:
            LiebherrBadRequestError: If invalid data is provided.
            LiebherrNotFoundError: If device is not reachable.
            LiebherrPreconditionFailedError: If device not onboarded.
            LiebherrConnectionError: If connection fails.
            LiebherrTimeoutError: If request times out.

        """
        await self._request(
            "POST",
            f"devices/{device_id}/controls/{CONTROL_PARTY_MODE}",
            json_data={"value": value},
        )

    async def set_night_mode(self, device_id: str, value: bool) -> None:
        """Set NightMode.

        Args:
            device_id: The device ID (serial number).
            value: True to enable, False to disable.

        Raises:
            LiebherrBadRequestError: If invalid data is provided.
            LiebherrNotFoundError: If device is not reachable.
            LiebherrPreconditionFailedError: If device not onboarded.
            LiebherrConnectionError: If connection fails.
            LiebherrTimeoutError: If request times out.

        """
        await self._request(
            "POST",
            f"devices/{device_id}/controls/{CONTROL_NIGHT_MODE}",
            json_data={"value": value},
        )

    async def set_presentation_light(self, device_id: str, target: int) -> None:
        """Set presentation light intensity.

        Args:
            device_id: The device ID (serial number).
            target: Light intensity value.

        Raises:
            LiebherrBadRequestError: If invalid data is provided.
            LiebherrNotFoundError: If device is not reachable.
            LiebherrPreconditionFailedError: If device not onboarded.
            LiebherrConnectionError: If connection fails.
            LiebherrTimeoutError: If request times out.

        """
        await self._request(
            "POST",
            f"devices/{device_id}/controls/{CONTROL_PRESENTATION_LIGHT}",
            json_data={"target": target},
        )

    # Special controls

    async def set_ice_maker(
        self, device_id: str, zone_id: int, mode: IceMakerMode
    ) -> None:
        """Set ice maker mode.

        Args:
            device_id: The device ID (serial number).
            zone_id: The zone ID.
            mode: Ice maker mode (OFF, ON, MAX_ICE).

        Raises:
            LiebherrBadRequestError: If invalid data is provided.
            LiebherrNotFoundError: If device is not reachable.
            LiebherrPreconditionFailedError: If device not onboarded.
            LiebherrUnsupportedError: If MaxIce not supported.
            LiebherrConnectionError: If connection fails.
            LiebherrTimeoutError: If request times out.

        """
        await self._request(
            "POST",
            f"devices/{device_id}/controls/{CONTROL_ICE_MAKER}",
            json_data={"zoneId": zone_id, "iceMakerMode": mode.value},
        )

    async def set_hydro_breeze(
        self, device_id: str, zone_id: int, mode: HydroBreezeMode
    ) -> None:
        """Set HydroBreeze mode.

        Args:
            device_id: The device ID (serial number).
            zone_id: The zone ID.
            mode: HydroBreeze mode (OFF, LOW, MEDIUM, HIGH).

        Raises:
            LiebherrBadRequestError: If invalid data is provided.
            LiebherrNotFoundError: If device is not reachable.
            LiebherrPreconditionFailedError: If device not onboarded.
            LiebherrUnsupportedError: If BioFreshPlus not enabled.
            LiebherrConnectionError: If connection fails.
            LiebherrTimeoutError: If request times out.

        """
        await self._request(
            "POST",
            f"devices/{device_id}/controls/{CONTROL_HYDRO_BREEZE}",
            json_data={"zoneId": zone_id, "hydroBreezeMode": mode.value},
        )

    async def set_bio_fresh_plus(
        self, device_id: str, zone_id: int, mode: BioFreshPlusMode
    ) -> None:
        """Set BioFreshPlus mode.

        Args:
            device_id: The device ID (serial number).
            zone_id: The zone ID.
            mode: BioFreshPlus mode.

        Raises:
            LiebherrBadRequestError: If invalid data is provided.
            LiebherrNotFoundError: If device is not reachable.
            LiebherrPreconditionFailedError: If device not onboarded.
            LiebherrUnsupportedError: If BioFreshPlus not enabled.
            LiebherrConnectionError: If connection fails.
            LiebherrTimeoutError: If request times out.

        """
        await self._request(
            "POST",
            f"devices/{device_id}/controls/{CONTROL_BIO_FRESH_PLUS}",
            json_data={"zoneId": zone_id, "bioFreshPlusMode": mode.value},
        )

    async def trigger_auto_door(
        self, device_id: str, zone_id: int, value: bool
    ) -> None:
        """Open or close auto door.

        Args:
            device_id: The device ID (serial number).
            zone_id: The zone ID.
            value: True to open, False to close.

        Raises:
            LiebherrBadRequestError: If invalid data is provided.
            LiebherrNotFoundError: If device is not reachable.
            LiebherrPreconditionFailedError: If device not onboarded.
            LiebherrUnsupportedError: If Auto Door not enabled.
            LiebherrConnectionError: If connection fails.
            LiebherrTimeoutError: If request times out.

        """
        await self._request(
            "POST",
            f"devices/{device_id}/controls/{CONTROL_AUTO_DOOR}",
            json_data={"zoneId": zone_id, "value": value},
        )

    # Convenience methods

    async def get_device_state(self, device_id: str) -> DeviceState:
        """Get complete device state (device info + all controls).

        Args:
            device_id: The device ID (serial number).

        Returns:
            DeviceState object containing device info and all controls.

        Raises:
            LiebherrNotFoundError: If device is not found.
            LiebherrConnectionError: If connection fails.
            LiebherrTimeoutError: If request times out.

        """
        device = await self.get_device(device_id)
        controls = await self.get_controls(device_id)
        return DeviceState(device=device, controls=controls)

    async def refresh_device(self, device_id: str) -> DeviceState:
        """Refresh and return current device state.

        This is an alias for get_device_state for better naming consistency.

        Args:
            device_id: The device ID (serial number).

        Returns:
            DeviceState object containing device info and all controls.

        Raises:
            LiebherrNotFoundError: If device is not found.
            LiebherrConnectionError: If connection fails.
            LiebherrTimeoutError: If request times out.

        """
        return await self.get_device_state(device_id)

    async def get_temperature_controls(
        self, device_id: str
    ) -> list[TemperatureControl]:
        """Get only temperature controls for a device.

        Args:
            device_id: The device ID (serial number).

        Returns:
            List of temperature controls.

        Raises:
            LiebherrNotFoundError: If device is not found.
            LiebherrConnectionError: If connection fails.
            LiebherrTimeoutError: If request times out.

        """
        controls = await self.get_controls(device_id)
        return [
            control for control in controls if isinstance(control, TemperatureControl)
        ]



"""Tests for Liebherr models."""

from typing import Any

import pytest

from pyliebherrhomeapi import (
    AutoDoorControl,
    BioFreshPlusControl,
    BioFreshPlusMode,
    Device,
    DeviceControl,
    DeviceState,
    DeviceType,
    DoorState,
    HydroBreezeControl,
    HydroBreezeMode,
    IceMakerControl,
    IceMakerMode,
    TemperatureControl,
    TemperatureUnit,
    ToggleControl,
    ZonePosition,
)
from pyliebherrhomeapi.models import parse_control


class TestDevice:
    """Tests for Device model."""

    def test_device_from_dict(self) -> None:
        """Test creating Device from dict."""
        data = {
            "deviceId": "12345",
            "nickname": "Kitchen Fridge",
            "deviceType": "FRIDGE",
            "imageUrl": "http://example.com/image.png",
            "deviceName": "Test Device",
        }
        device = Device.from_dict(data)
        assert device.device_id == "12345"
        assert device.nickname == "Kitchen Fridge"
        assert device.device_type == DeviceType.FRIDGE
        assert device.image_url == "http://example.com/image.png"
        assert device.device_name == "Test Device"

    def test_device_from_dict_minimal(self) -> None:
        """Test creating Device from dict with minimal data."""
        data = {"deviceId": "12345"}
        device = Device.from_dict(data)
        assert device.device_id == "12345"
        assert device.nickname is None
        assert device.device_type is None
        assert device.image_url is None
        assert device.device_name is None

    @pytest.mark.parametrize(
        (
            "device_type",
            "expected_is_fridge",
            "expected_is_freezer",
            "expected_is_combi",
            "expected_is_wine",
        ),
        [
            (DeviceType.FRIDGE, True, False, False, False),
            (DeviceType.FREEZER, False, True, False, False),
            (DeviceType.COMBI, False, False, True, False),
            (DeviceType.WINE, False, False, False, True),
        ],
    )
    def test_device_type_methods(
        self,
        device_type: DeviceType,
        expected_is_fridge: bool,
        expected_is_freezer: bool,
        expected_is_combi: bool,
        expected_is_wine: bool,
    ) -> None:
        """Test device type checking methods."""
        device = Device(device_id="1", device_type=device_type)
        assert device.is_fridge() is expected_is_fridge
        assert device.is_freezer() is expected_is_freezer
        assert device.is_combi() is expected_is_combi
        assert device.is_wine() is expected_is_wine

    def test_is_methods_with_none_type(self) -> None:
        """Test is_* methods when device_type is None."""
        device = Device(device_id="1", device_type=None)
        assert device.is_fridge() is False
        assert device.is_freezer() is False
        assert device.is_combi() is False
        assert device.is_wine() is False

    def test_device_unknown_type_retains_raw_value(self) -> None:
        """Unknown device types are preserved as raw strings."""

        device = Device.from_dict({"deviceId": "1", "deviceType": "NEW_TYPE"})

        assert device.device_type == "NEW_TYPE"
        assert device.is_fridge() is False


class TestTemperatureControl:
    """Tests for TemperatureControl model."""

    def test_temperature_control_from_dict(self) -> None:
        """Test creating TemperatureControl from dict."""
        data = {
            "name": "temperature",
            "type": "TemperatureControl",
            "zoneId": 0,
            "zonePosition": "top",
            "value": 4,
            "target": 5,
            "min": 2,
            "max": 8,
            "unit": "°C",
        }
        control = TemperatureControl.from_dict(data)
        assert control.name == "temperature"
        assert control.type == "TemperatureControl"
        assert control.zone_id == 0
        assert control.zone_position == ZonePosition.TOP
        assert control.value == 4
        assert control.target == 5
        assert control.min == 2
        assert control.max == 8
        assert control.unit == TemperatureUnit.CELSIUS

    def test_temperature_control_from_dict_minimal(self) -> None:
        """Test creating TemperatureControl from dict with minimal data."""
        data = {"name": "temperature", "type": "TemperatureControl", "zoneId": 0}
        control = TemperatureControl.from_dict(data)
        assert control.name == "temperature"
        assert control.zone_id == 0
        assert control.value is None
        assert control.target is None

    def test_temperature_control_unknown_zone_position(self) -> None:
        """Unknown zone positions are preserved as raw strings."""

        data = {
            "name": "temperature",
            "type": "TemperatureControl",
            "zoneId": 0,
            "zonePosition": "mystery",
            "unit": "KELVIN",
        }

        control = TemperatureControl.from_dict(data)

        assert control.zone_position == "mystery"
        assert control.unit == "KELVIN"

    @pytest.mark.parametrize(
        ("min_temp", "max_temp", "test_value", "expected"),
        [
            # Within range
            (2, 8, 4, True),
            (2, 8, 2, True),
            (2, 8, 8, True),
            # Below min
            (2, 8, 1, False),
            (2, 8, 0, False),
            # Above max
            (2, 8, 9, False),
            (2, 8, 10, False),
            # No limits
            (None, None, -100, True),
            (None, None, 100, True),
            # Only min
            (2, None, 2, True),
            (2, None, 100, True),
            (2, None, 1, False),
            # Only max
            (None, 8, 8, True),
            (None, 8, -100, True),
            (None, 8, 9, False),
        ],
    )
    def test_validate_temperature(
        self,
        min_temp: int | None,
        max_temp: int | None,
        test_value: int,
        expected: bool,
    ) -> None:
        """Test validate_temperature with various min/max combinations."""
        control = TemperatureControl(
            name="temperature",
            type="TemperatureControl",
            zone_id=0,
            min=min_temp,
            max=max_temp,
        )
        assert control.validate_temperature(test_value) is expected


class TestToggleControl:
    """Tests for ToggleControl model."""

    def test_toggle_control_from_dict(self) -> None:
        """Test creating ToggleControl from dict."""
        data = {
            "name": "superfrost",
            "type": "ToggleControl",
            "zoneId": 0,
            "zonePosition": "bottom",
            "value": True,
        }
        control = ToggleControl.from_dict(data)
        assert control.name == "superfrost"
        assert control.type == "ToggleControl"
        assert control.zone_id == 0
        assert control.zone_position == ZonePosition.BOTTOM
        assert control.value is True

    def test_toggle_control_from_dict_minimal(self) -> None:
        """Test creating ToggleControl from dict with minimal data."""
        data = {"name": "superfrost", "type": "ToggleControl"}
        control = ToggleControl.from_dict(data)
        assert control.name == "superfrost"
        assert control.zone_id is None
        assert control.value is None


class TestAutoDoorControl:
    """Tests for AutoDoorControl model."""

    def test_auto_door_control_from_dict(self) -> None:
        """Test creating AutoDoorControl from dict."""
        data = {
            "name": "autodoor",
            "type": "AutoDoorControl",
            "zoneId": 0,
            "zonePosition": "middle",
            "value": "CLOSED",
        }
        control = AutoDoorControl.from_dict(data)
        assert control.name == "autodoor"
        assert control.type == "AutoDoorControl"
        assert control.zone_id == 0
        assert control.zone_position == ZonePosition.MIDDLE
        assert control.value == DoorState.CLOSED

    def test_auto_door_control_from_dict_minimal(self) -> None:
        """Test creating AutoDoorControl from dict with minimal data."""
        data = {"name": "autodoor", "type": "AutoDoorControl", "zoneId": 0}
        control = AutoDoorControl.from_dict(data)
        assert control.name == "autodoor"
        assert control.zone_id == 0
        assert control.value is None


class TestIceMakerControl:
    """Tests for IceMakerControl model."""

    def test_ice_maker_control_from_dict(self) -> None:
        """Test creating IceMakerControl from dict."""
        data = {
            "name": "icemaker",
            "type": "IceMakerControl",
            "zoneId": 0,
            "zonePosition": "top",
            "iceMakerMode": "MAX_ICE",
            "hasMaxIce": True,
        }
        control = IceMakerControl.from_dict(data)
        assert control.name == "icemaker"
        assert control.type == "IceMakerControl"
        assert control.zone_id == 0
        assert control.zone_position == ZonePosition.TOP
        assert control.ice_maker_mode == IceMakerMode.MAX_ICE
        assert control.has_max_ice is True

    def test_ice_maker_control_from_dict_minimal(self) -> None:
        """Test creating IceMakerControl from dict with minimal data."""
        data = {"name": "icemaker", "type": "IceMakerControl", "zoneId": 0}
        control = IceMakerControl.from_dict(data)
        assert control.name == "icemaker"
        assert control.zone_id == 0
        assert control.ice_maker_mode is None
        assert control.has_max_ice is None


class TestHydroBreezeControl:
    """Tests for HydroBreezeControl model."""

    def test_hydro_breeze_control_from_dict(self) -> None:
        """Test creating HydroBreezeControl from dict."""
        data = {
            "name": "hydrobreeze",
            "type": "HydroBreezeControl",
            "zoneId": 0,
            "currentMode": "HIGH",
        }
        control = HydroBreezeControl.from_dict(data)
        assert control.name == "hydrobreeze"
        assert control.type == "HydroBreezeControl"
        assert control.zone_id == 0
        assert control.current_mode == HydroBreezeMode.HIGH

    def test_hydro_breeze_control_from_dict_minimal(self) -> None:
        """Test creating HydroBreezeControl from dict with minimal data."""
        data = {"name": "hydrobreeze", "type": "HydroBreezeControl", "zoneId": 0}
        control = HydroBreezeControl.from_dict(data)
        assert control.name == "hydrobreeze"
        assert control.zone_id == 0
        assert control.current_mode is None


class TestBioFreshPlusControl:
    """Tests for BioFreshPlusControl model."""

    def test_biofresh_plus_control_from_dict(self) -> None:
        """Test creating BioFreshPlusControl from dict."""
        data = {
            "name": "biofreshplus",
            "type": "BioFreshPlusControl",
            "zoneId": 0,
            "currentMode": "ZERO_ZERO",
            "supportedModes": ["ZERO_ZERO", "ZERO_MINUS_TWO"],
            "temperatureUnit": "°C",
        }
        control = BioFreshPlusControl.from_dict(data)
        assert control.name == "biofreshplus"
        assert control.type == "BioFreshPlusControl"
        assert control.zone_id == 0
        assert control.current_mode == BioFreshPlusMode.ZERO_ZERO
        assert len(control.supported_modes) == 2
        assert BioFreshPlusMode.ZERO_ZERO in control.supported_modes
        assert control.temperature_unit == TemperatureUnit.CELSIUS

    def test_biofresh_plus_control_from_dict_minimal(self) -> None:
        """Test creating BioFreshPlusControl from dict with minimal data."""
        data = {"name": "biofreshplus", "type": "BioFreshPlusControl", "zoneId": 0}
        control = BioFreshPlusControl.from_dict(data)
        assert control.name == "biofreshplus"
        assert control.zone_id == 0
        assert control.current_mode is None
        assert len(control.supported_modes) == 0
        assert control.temperature_unit is None


class TestParseControl:
    """Tests for parse_control function."""

    @pytest.mark.parametrize(
        ("control_type", "expected_class"),
        [
            ("TemperatureControl", TemperatureControl),
            ("ToggleControl", ToggleControl),
            ("AutoDoorControl", AutoDoorControl),
            ("IceMakerControl", IceMakerControl),
            ("HydroBreezeControl", HydroBreezeControl),
            ("BioFreshPlusControl", BioFreshPlusControl),
            ("UnknownControl", ToggleControl),  # Fallback
        ],
    )
    def test_parse_control_types(
        self, control_type: str, expected_class: type[DeviceControl]
    ) -> None:
        """Test parsing various control types."""
        data: dict[str, Any] = {"name": "test", "type": control_type}
        if control_type != "ToggleControl":
            data["zoneId"] = 0
        control = parse_control(data)
        assert isinstance(control, expected_class)


class TestDeviceState:
    """Tests for DeviceState model."""

    @pytest.fixture
    def sample_device(self) -> Device:
        """Create a sample device."""
        return Device(device_id="12345", device_type=DeviceType.FRIDGE)

    @pytest.fixture
    def sample_controls(self) -> list[DeviceControl]:
        """Create sample controls."""
        return [
            TemperatureControl(name="temp1", type="TemperatureControl", zone_id=0),
            TemperatureControl(name="temp2", type="TemperatureControl", zone_id=1),
            ToggleControl(name="toggle1", type="ToggleControl", zone_id=0),
            ToggleControl(name="toggle2", type="ToggleControl", zone_id=1),
            AutoDoorControl(name="door1", type="AutoDoorControl", zone_id=0),
            IceMakerControl(name="ice1", type="IceMakerControl", zone_id=0),
            HydroBreezeControl(name="hydro1", type="HydroBreezeControl", zone_id=1),
            BioFreshPlusControl(name="bio1", type="BioFreshPlusControl", zone_id=1),
        ]

    def test_device_state_creation(
        self, sample_device: Device, sample_controls: list[DeviceControl]
    ) -> None:
        """Test creating DeviceState."""
        state = DeviceState(device=sample_device, controls=sample_controls)
        assert state.device == sample_device
        assert len(state.controls) == 8

    def test_device_state_default_controls(self, sample_device: Device) -> None:
        """Test DeviceState with default empty controls."""
        state = DeviceState(device=sample_device)
        assert state.device == sample_device
        assert len(state.controls) == 0

    @pytest.mark.parametrize(
        ("method_name", "expected_type", "expected_count"),
        [
            ("get_temperature_controls", TemperatureControl, 2),
            ("get_toggle_controls", ToggleControl, 2),
            ("get_auto_door_controls", AutoDoorControl, 1),
            ("get_ice_maker_controls", IceMakerControl, 1),
            ("get_hydro_breeze_controls", HydroBreezeControl, 1),
            ("get_biofresh_plus_controls", BioFreshPlusControl, 1),
        ],
    )
    def test_get_control_type_methods(
        self,
        sample_device: Device,
        sample_controls: list[DeviceControl],
        method_name: str,
        expected_type: type[DeviceControl],
        expected_count: int,
    ) -> None:
        """Test various get_*_controls methods."""
        state = DeviceState(device=sample_device, controls=sample_controls)
        method = getattr(state, method_name)
        controls = method()
        assert len(controls) == expected_count
        assert all(isinstance(c, expected_type) for c in controls)

    def test_get_control_by_name_found(
        self, sample_device: Device, sample_controls: list[DeviceControl]
    ) -> None:
        """Test get_control_by_name when control exists."""
        state = DeviceState(device=sample_device, controls=sample_controls)
        control = state.get_control_by_name("temp1")
        assert control is not None
        assert control.name == "temp1"

    def test_get_control_by_name_not_found(
        self, sample_device: Device, sample_controls: list[DeviceControl]
    ) -> None:
        """Test get_control_by_name when control doesn't exist."""
        state = DeviceState(device=sample_device, controls=sample_controls)
        control = state.get_control_by_name("nonexistent")
        assert control is None

    @pytest.mark.parametrize(
        ("zone_id", "expected_count"),
        [
            (0, 4),  # temp1, toggle1, door1, ice1
            (1, 4),  # temp2, toggle2, hydro1, bio1
        ],
    )
    def test_get_controls_by_zone(
        self,
        sample_device: Device,
        sample_controls: list[DeviceControl],
        zone_id: int,
        expected_count: int,
    ) -> None:
        """Test get_controls_by_zone method."""
        state = DeviceState(device=sample_device, controls=sample_controls)
        zone_controls = state.get_controls_by_zone(zone_id)
        assert len(zone_controls) == expected_count

    def test_get_controls_by_zone_empty(self, sample_device: Device) -> None:
        """Test get_controls_by_zone with no controls."""
        state = DeviceState(device=sample_device, controls=[])
        zone_controls = state.get_controls_by_zone(0)
        assert len(zone_controls) == 0

    def test_filter_methods_with_empty_controls(self, sample_device: Device) -> None:
        """Test all filter methods with empty controls."""
        state = DeviceState(device=sample_device, controls=[])
        assert len(state.get_temperature_controls()) == 0
        assert len(state.get_toggle_controls()) == 0
        assert len(state.get_auto_door_controls()) == 0
        assert len(state.get_ice_maker_controls()) == 0
        assert len(state.get_hydro_breeze_controls()) == 0
        assert len(state.get_biofresh_plus_controls()) == 0



"""Exceptions for pyliebherrhomeapi."""


class LiebherrError(Exception):
    """Base exception for Liebherr API errors."""


class LiebherrConnectionError(LiebherrError):
    """Exception raised when connection to Liebherr API fails."""


class LiebherrTimeoutError(LiebherrError):
    """Exception raised when request to Liebherr API times out."""


class LiebherrAuthenticationError(LiebherrError):
    """Exception raised when authentication fails."""


class LiebherrBadRequestError(LiebherrError):
    """Exception raised when invalid data is provided."""


class LiebherrNotFoundError(LiebherrError):
    """Exception raised when device is not reachable."""


class LiebherrPreconditionFailedError(LiebherrError):
    """Exception raised when precondition fails (device not onboarded)."""


class LiebherrUnsupportedError(LiebherrError):
    """Exception raised when operation is not supported."""


class LiebherrServerError(LiebherrError):
    """Exception raised when server returns 500 error."""
